#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <SDL/SDL_endian.h>

#include "fat.h"
#include "fat_internal.h"
#include "sectorbits.h"
#include "spinner.h"
#include "mydebug.h"

const char *fat_type_str[];

static void FAT_procflags(char *buf, uint8_t flags);
static uint32_t FAT_table_length(FAT *fat);
static cluster_e FAT_readtable(FAT *fat, int index);

#define CLUSTER_TO_SECTOR(FAT,CLUSTER) ((CLUSTER)<<((FAT)->clusterbits))
#define SECTOR_TO_BYTES(FAT,SECTOR)    (((uint64_t)(SECTOR))<<((FAT)->sectorbits))

int write_sector(FAT *fat, int fdout, int num)
{
  unsigned const char *buf=(unsigned char *)(fat->clusterbuf);
  int size=num<<(fat->sectorbits);
  while(size > 0)
  {
    int retval=write(fdout,buf,size);

    if(retval<=0)
    switch(errno)
    {
    int extra;

    case EINTR:
    case EAGAIN:
      continue;

    case EBADF:
    case EINVAL:
      perror("Could not write sector");
      return(-1);

    case EFBIG:
    case ENOSPC:
      perror("Exceeded file length limit");
      return(-1);

    case EIO:
      perror("Bad sector?  Skipping...");
      extra=512-(size%512);
      size-=extra;
      buf+=extra;
      continue;

    default:
      perror("Unknown write error");
      return(-1);
    }

    buf +=retval;
    size-=retval;
  }
  return(0);
}

int read_sector(FAT *fat, int num)
{
  int len=num<<(fat->sectorbits);// + fat->clusterbits);
  unsigned char *buf=(unsigned char *)(fat->clusterbuf);

  while(len>0)
  {
    ssize_t retval=read(fat->fd,buf,len);

    if(retval <= 0)
    switch(errno)
    {
    int extra;
    case EAGAIN:
      continue;

    case EBADF:
    case EINTR:
      perror("Read error");
      return(-1);

    case EFAULT:
      perror("Read error, bad address");
      return(-1);

    case EIO:
      perror("Bad sector?  skipping");
      extra=512-(len%512);
      memset(buf,0,extra);
      len-=extra;
      buf+=extra;
      continue;

    default:
      perror("Unknown error");
      return(-1);
    }

    len-=retval;
    buf+=retval;    
  }

  return(0);
}

#define READ_CLUSTER(f) read_sector((f),1<<((f)->clusterbits))
#define WRITE_CLUSTER(f,fd) write_sector((f),(fd),1<<((f)->clusterbits))
#define SEEK_CLUSTER(f,fd,c) seek_sector((f),(fd),CLUSTER_TO_SECTOR((f),(c))+(f)->datastart)

int seek_sector(FAT *fat, int fdout, sector_t sector)
{
  pos_t offset=SECTOR_TO_BYTES(fat,sector);
  DEBUG_PRINT(("Seeking 0x%016lx\n",offset));
  if(lseek64(fdout,offset,SEEK_SET)!=offset)
    return(-1);
  if(lseek64(fat->fd,offset,SEEK_SET)!=offset)
    return(-1);
  return(0);
}

int copy_clusters(FAT *fat, int fdout, cluster_t start, int len)
{
  if(SEEK_CLUSTER(fat,fdout,start)<0)
    return(-1);

  while(len>0)
  {
    if(READ_CLUSTER(fat)<0)
      return(-1);
    if(WRITE_CLUSTER(fat,fdout)<0)
      return(-1);

    len--;
  }

  return(0);
}


int FAT_copy(FAT *fat, int fdout)
{
  int index;
  int32_t first_cluster=-1;
  uint32_t cluster_len =0;

  INFO_PRINT(("Copying volume..."));

  lseek(fdout,0,SEEK_SET);
  {
    pos_t pos=(pos_t)0;
    pos_t datapos=SECTOR_TO_BYTES(fat,fat->datastart);
    ssize_t retval=0;
    INFO_PRINT(("\n\tCopying boot sector, fat tables(%u K)...",datapos>>10));

    spin_start();
      while((retval >= 0)&&(pos < datapos))
      {
        retval=write(fdout,((char *)fat->boot)+pos,datapos-pos);
        if(retval > 0) pos+=(pos_t)retval;
      }
    spin_stop();

    INFO_PRINT(("Done.\n"));
  }

  INFO_PRINT(("\tWriting last sector..."));
  spin_start();
    seek_sector(fat,fdout,BOOT_total_sectors(fat->boot)-1);
    read_sector(fat,1);
    write_sector(fat,fdout,1);
  spin_stop();
  INFO_PRINT(("Done.\n" "\tCopying used clusters..."));


  spin_start();
    for(index=0; index<fat->tablesize; index++)
    switch(FAT_readtable(fat,index))
    {
    case CLUSTER_RESERVED:
    case CLUSTER_LAST:
    case CLUSTER_USER:
      if(first_cluster < 0)
      {
        first_cluster=index;
        cluster_len=1;
      }
      else if(index==(first_cluster+cluster_len))
        cluster_len++;
      else
      {
        copy_clusters(fat,fdout,first_cluster,cluster_len);
        first_cluster=index;
        cluster_len=1;
      }
      break;
    case CLUSTER_BAD:
      break;
    default:
      break;
    }

  if(first_cluster >= 0)
    copy_clusters(fat,fdout,first_cluster,cluster_len);
  spin_stop();

  INFO_PRINT(("Done.\n"));
}

fat_e FAT_type(FAT *fat)
{
  return(fat->type);
}

FAT *FAT_open(const char *fname)
{
  FAT *fat=(FAT *)malloc(sizeof(FAT));
  int n,mapsize=0;

  if(fat == NULL)
    goto FAT_OPEN_ERR;

  fat->fd=open(fname,O_RDONLY);
  if(fat->fd < 0)
    goto FAT_OPEN_ERR;

  fat->pagesize=getpagesize();
  fat->boot=(FAT_boot *)mmap(NULL,fat->pagesize,PROT_READ,MAP_SHARED,fat->fd,0);
  if(fat->boot == MAP_FAILED)
    goto FAT_OPEN_ERR;

  mapsize=fat->pagesize;

  if(BOOT_valid(fat->boot)<0)
    goto FAT_OPEN_ERR;

  fat->type=BOOT_fstype(fat->boot);
  if(fat->type<0)
  {
    ERROR_PRINT(("Not a FAT filesystem\n"));
    goto FAT_OPEN_ERR;
  }

  fat->sectorbits=sector_bits(BOOT_BYTES_PER_SECTOR(fat->boot));
  fat->clusterbits=sector_bits(fat->boot->sectors_per_cluster);
  fat->datastart=BOOT_data_start(fat->boot);
  fat->clusterbuf=malloc(BOOT_CLUSTERSIZE(fat->boot));

  ERROR_PRINT(("Is a valid %s filesystem\n",fat_type_str[fat->type]));
  INFO_PRINT(("\t%d bytes per sector\n",BOOT_BYTES_PER_SECTOR(fat->boot)));
  INFO_PRINT(("\t%d bytes per cluster\n",BOOT_CLUSTERSIZE(fat->boot)));
  INFO_PRINT(("\t%d sectors\n",BOOT_total_sectors(fat->boot)));
  INFO_PRINT(("\t%d clusters\n",BOOT_clusters(fat->boot)));
  INFO_PRINT(("\t%d root sectors\n",BOOT_root_sectors(fat->boot)));
  INFO_PRINT(("\tTable start is\t0x%08x\n",
    SECTOR_TO_BYTES(fat,BOOT_fat_table_start(fat->boot))));
  INFO_PRINT(("\tRoot start is\t0x%08x\n",
    SECTOR_TO_BYTES(fat,BOOT_root_start(fat->boot))));
  INFO_PRINT(("\tData start is\t0x%08x\n",
    SECTOR_TO_BYTES(fat,BOOT_data_start(fat->boot))));

  mapsize=SECTOR_TO_BYTES(fat,BOOT_data_start(fat->boot));
  munmap(fat->boot,fat->pagesize);
  fat->boot=mmap(NULL,mapsize,PROT_READ,MAP_SHARED,fat->fd,0);
  if(fat->boot == MAP_FAILED)
  {
    INFO_PRINT(("Couldn't map\n"));
    goto FAT_OPEN_ERR;
  }

  DEBUG_PRINT(("\tRemapped to 0x%08x bytes\n",mapsize));

  fat->table =(fat_table *)(((char *)fat->boot)+
    SECTOR_TO_BYTES(fat,BOOT_fat_table_start(fat->boot)));
  fat->tablesize=FAT_table_length(fat);
  return(fat);

FAT_OPEN_ERR:
  if(fat != NULL)
  {
    if(fat->clusterbuf != NULL) free(fat->clusterbuf);
    if(fat->boot != MAP_FAILED) munmap(fat->boot,mapsize);
    if(fat->fd >= 0) close(fat->fd);
    free(fat);
  }
  return(NULL);
}

void FAT_close(FAT *fat)
{
  if(fat != NULL)
  {
    uint32_t size=SECTOR_TO_BYTES(fat,BOOT_data_start(fat->boot));
    DEBUG_PRINT(("Unmapping %d bytes, starting at %p\n",size,fat));
    munmap(fat->boot,size);
    if(fat->fd >= 0) close(fat->fd);
    if(fat->clusterbuf != NULL) free(fat->clusterbuf);
    free(fat);
  }
}

static void FAT_procflags(char *buf, uint8_t flags)
{
  int n;
  struct
  {
    fat_attr_t flag;
    char present,absent;
  } flagdesc[]={ {FAT_ATTR_LFN      , 'L', '-'},
                 {FAT_ATTR_READONLY , 'R', '-'},
                 {FAT_ATTR_SYSTEM   , 'S', '-'},
                 {FAT_ATTR_HIDDEN   , 'H', '-'},
                 {FAT_ATTR_VOLUME   , 'V', '-'},
                 {FAT_ATTR_DIRECTORY, 'D', '-'},
                 {                0,'\0','\0'}};

  for(n=0; flagdesc[n].flag != 0; n++)
  {
    if((flags & flagdesc[n].flag)==flagdesc[n].flag)
    {
      buf[n]=flagdesc[n].present;
      flags &= ~flagdesc[n].flag;
    }
    else
      buf[n]=flagdesc[n].absent;
  }

  buf[n]='\0';
}

static uint32_t FAT_table_length(FAT *fat)
{
  int size=BOOT_clusters(fat->boot);
  switch(fat->type)
  {
  case FAT_12:
    return((size*2)/3);
  case FAT_16:
    return(size/2);
  case FAT_32:
    return(size/4);
  default:
    return(0);
  }
}

const char *fat_type_str[]={"FAT12","FAT16","FAT32",NULL};

static cluster_e FAT_readtable(FAT *fat, int index)
{
  switch(fat->type)
  {
  uint8_t which;

  uint32_t val;
  case FAT_32:
    // Microsoft sez highest nibble is reserved
    val=SDL_SwapLE32(fat->table->fat32[index])&0x0fffffff;
    if((val >= FAT32_END1)&&(val<=FAT32_END2))
      return(CLUSTER_LAST);

    switch(val)
    {
    case FAT32_AVAILABLE: return(CLUSTER_AVAILABLE);
    case FAT32_RESERVED:  return(CLUSTER_RESERVED);
    case FAT32_BAD:       return(CLUSTER_BAD);
    default:              return(CLUSTER_USER);
    }
    break;
  case FAT_16:
    val=SDL_SwapLE16(fat->table->fat16[index]);
    if((val >= FAT16_END1)&&(val<=FAT16_END2))
      return(CLUSTER_LAST);
    switch(val)
    {
    case FAT16_AVAILABLE: return(CLUSTER_AVAILABLE);
    case FAT16_RESERVED:  return(CLUSTER_RESERVED);
    case FAT16_BAD:       return(CLUSTER_BAD);
    default:
      return(CLUSTER_USER);
    }
    break;
  case FAT_12:
    which=index&0x07;
    index>>=3;
    switch(which)
    {
    case 0:
      val=SDL_SwapLE16(fat->table->fat12[index].fat12_0);
      break;
    case 1:
      val=SDL_SwapLE16(fat->table->fat12[index].fat12_1);
      break;
    case 2:
      val=SDL_SwapLE16(fat->table->fat12[index].fat12_2);
      break;
    case 3:
      val=SDL_SwapLE16(fat->table->fat12[index].fat12_3);
      break;
    case 4:
      val=SDL_SwapLE16(fat->table->fat12[index].fat12_4);
      break;
    case 5:
      val=SDL_SwapLE16(fat->table->fat12[index].fat12_5);
      break;
    case 6:
      val=SDL_SwapLE16(fat->table->fat12[index].fat12_6);
      break;
    case 7:
      val=SDL_SwapLE16(fat->table->fat12[index].fat12_7);
      break;
    default:
      ERROR_PRINT(("WTF?  which>7 for FAT12\n"));
      abort();
      break;
    }
    if((val >= FAT12_END1)&&(val<=FAT12_END2))
      return(CLUSTER_LAST);

    switch(val)
    {
    case FAT12_AVAILABLE: return(CLUSTER_AVAILABLE);
    case FAT12_RESERVED:  return(CLUSTER_RESERVED);
    case FAT12_BAD:       return(CLUSTER_BAD);
    default:              return(CLUSTER_USER);
    }
  }
}
