#include "fat_boot.h"
#include "mydebug.h"

sector_t BOOT_data_start(const FAT_boot *fb)
{
  return(BOOT_RESERVED_SECTORS(fb) +
    (BOOT_fat_sectors(fb) * fb->num_fats) +
    BOOT_root_sectors(fb));
}

sector_t BOOT_fat_table_start(const FAT_boot *fb)
{
  return(BOOT_RESERVED_SECTORS(fb));
}

sector_t BOOT_root_start(const FAT_boot *fb)
{
  return(BOOT_RESERVED_SECTORS(fb) + (BOOT_FAT_SECTORS(fb)*fb->num_fats));
}

sector_t BOOT_fat_sectors(const FAT_boot *fb)
{
  if(BOOT_FAT_SECTORS(fb) != 0)
    return(BOOT_FAT_SECTORS(fb));
  else
    return(BOOT32_FAT_SECTORS(fb));
}

sector_t BOOT_root_sectors(const FAT_boot *fb)
{
 return((((BOOT_ROOT_ENTRIES(fb)+1)*sizeof(FAT_direntry))-1) /
    BOOT_BYTES_PER_SECTOR(fb));
}

sector_t BOOT_total_sectors(const FAT_boot *fb)
{
  if(BOOT_SECTORS(fb) != 0) return(BOOT_SECTORS(fb));
  else                      return(BOOT_SECTORS32(fb));
}

cluster_t BOOT_clusters(const FAT_boot *fb)
{
  sector_t sectors=BOOT_total_sectors(fb) - 
    ( BOOT_RESERVED_SECTORS(fb) + 
      (BOOT_fat_sectors(fb) * fb->num_fats) +
      BOOT_root_sectors(fb)                   );
  return(sectors/(fb->sectors_per_cluster));
}

fat_e BOOT_fstype(const FAT_boot *fb)
{
  uint32_t clustercount=BOOT_clusters(fb);
  if(clustercount == 0)
    return(FAT_INVALID);
  else if(clustercount < 4085)
    return(FAT_12);
  else if(clustercount < 65525)
    return(FAT_16);
  else
    return(FAT_32);
}

void print_boot_signiature(const FAT_bootsig sig)
{
  INFO_PRINT(("Extended boot signiature:\n"));
  INFO_PRINT(("\tVolume #:     0x%08x\n",SDL_SwapLE32(sig.volumeid)));
  INFO_PRINT(("\tVolume Label: %.11s\n",sig.volumelabel));
  INFO_PRINT(("\tSystem Type:  %.8s\n",sig.systype));
}

int BOOT_valid(const FAT_boot *fb)
{
  int expect_fat32=0;
  int n;

  if(sizeof(FAT_boot) != 512)
  {
    ERROR_PRINT(("Compilation error!  Boot sector should be 512, not %d\n",
      sizeof(FAT_boot)));
    return(-1);
  }

  if((fb->hex_55 != 0x55)||(fb->hex_aa != 0xaa))
  {
    ERROR_PRINT(("Bad end marker 0x%02x 0x%02x\n",fb->hex_55,fb->hex_aa));
    return(-1);
  }


  if(!(((fb->jmpboot[0] == 0xeb) && (fb->jmpboot[2] == 0x90))
    || (fb->jmpboot[0] == 0xe9)))
  {
    ERROR_PRINT(("Invalid boot jump 0x%02x 0x%02x 0x%02x\n",
      fb->jmpboot[0],fb->jmpboot[1],fb->jmpboot[2]));
    return(-1);
  }

  if(strncmp(BOOT_OEMNAME_DEFAULT,fb->oemname,8)!=0)
    INFO_PRINT(("Warning:  OEMname is \"%8s\", not \"%8s\"\n",
      fb->oemname,BOOT_OEMNAME_DEFAULT));


  switch(BOOT_BYTES_PER_SECTOR(fb))
  {
  case 1024:
  case 2048:
  case 4096:
    INFO_PRINT(("Warning:  Uncommon (but valid) bytes-per-sector %d\n",
      BOOT_BYTES_PER_SECTOR(fb)));
  case 512:
    break;

  default:
    ERROR_PRINT(("Invalid bytes-per-sector %d\n",
      BOOT_BYTES_PER_SECTOR(fb)));
    return(-1);
  }


  n=1;
  while(1)
  {
    if(n>256)
    {
      ERROR_PRINT(("Invalid sectors-per-cluster %d\n",fb->sectors_per_cluster));
      return(-1);
    }

    if(n == fb->sectors_per_cluster)
      break;

    n<<=1;
  }

  if(BOOT_RESERVED_SECTORS(fb) == 0)
  {
    ERROR_PRINT(("Invalid reserved-sector count %d\n",
      BOOT_RESERVED_SECTORS(fb)));
    return(-1);
  }
  else if(BOOT_RESERVED_SECTORS(fb) == 32) { }
  else if(BOOT_RESERVED_SECTORS(fb) != 1)
    INFO_PRINT(("Warning:  uncommon reserved-sector count %d\n",
      BOOT_RESERVED_SECTORS(fb)));

  if(fb->num_fats == 0)
  {
    ERROR_PRINT(("Invalid # of fats:  %d\n",fb->num_fats));
    return(-1);
  }
  else if(fb->num_fats != 2)
    INFO_PRINT(("Warning:  Uncommon # of fats %d\n",fb->num_fats));

  if(BOOT_ROOT_ENTRIES(fb) == 0)
    expect_fat32=1;
  else if((BOOT_ROOT_ENTRIES(fb)*32)%BOOT_BYTES_PER_SECTOR(fb))
  {
    ERROR_PRINT(("Invalid # of root entries: %d\n",BOOT_ROOT_ENTRIES(fb)));
    return(-1);
  }

  if(expect_fat32 && (BOOT_SECTORS(fb) != 0))
  {
    ERROR_PRINT(("FAT32 partition with 16-bit sector count?!  BARF\n"));
    return(-1);
  }

  switch(fb->media_type)
  {
//  case FAT_UNKNOWN1:
  case FAT_UNKNOWN2:
  case FAT_UNKNOWN3:
  case FAT_UNKNOWN4:
  case FAT_UNKNOWN5:
  case FAT_UNKNOWN6:
  case FAT_UNKNOWN7:
    INFO_PRINT(("Warning:  Nonstandard media type 0x%02x\n",fb->media_type));
    break;

  case FAT_MEDIA_FIXED:
    INFO_PRINT(("FAT media type 'fixed'\n"));
    break;

  case FAT_MEDIA_REMOVABLE:
    INFO_PRINT(("FAT media type 'removable'\n"));
    break;

  default:
    ERROR_PRINT(("Unheard of media type 0x%02x\n",fb->media_type));
    return(-1);
  }

  if(expect_fat32 && (BOOT_FAT_SECTORS(fb) != 0))
  {
    ERROR_PRINT(("Nonzero FAT size %d for FAT32!?  BARF\n",
      BOOT_FAT_SECTORS(fb)));
    return(-1);
  }

  if(BOOT_HIDDEN_SECTORS(fb) != 0)
    INFO_PRINT(("Warning:  Uncommon # of hidden sectors %d\n",
      BOOT_HIDDEN_SECTORS(fb)));

  if(expect_fat32)
  {
    if(BOOT_SECTORS32(fb)==0)
    {
      ERROR_PRINT(("Zero-sized FAT32 partition!  BARF\n"));
      return(-1);
    }
  }
  else if((BOOT_SECTORS32(fb)>0)&&(BOOT_SECTORS32(fb) < 0x10000))
  {
    ERROR_PRINT(("Bad fat12/fat16 32-bit size 0x%08x\n",
      BOOT_SECTORS32(fb)));
    return(-1);
  }

  if(expect_fat32)
  {
    if(BOOT32_MIRRORED(fb))
      INFO_PRINT(("Changes mirrored into all %d fats\n",
        fb->num_fats));
    else
      INFO_PRINT(("Changes only put in fat %d/%d\n",
      BOOT32_ACTIVE_FAT(fb)+1,fb->num_fats));

    if(BOOT32_VERSION(fb) != 0x0000)
    {
      ERROR_PRINT(("Unknown FAT32 version 0x%04x\n",
        BOOT32_VERSION(fb)));
      return(-1);
    }

    if(BOOT_HAVE_BOOTSIG(fb->fat32))
    {
//      fprintf(stderr,"Volume %: %d\n",SDL_SwapLE32(fb->fat32.sig.volumeid));
//      fprintf(stderr,"label: %.11s\n",fb->fat32.sig.volumelabel);
//      fprintf(stderr,"systype: %.8s\n",sig.systype);
      print_boot_signiature(fb->fat32.bootsig);
    }
    else
      INFO_PRINT(("No boot signiature.\n"));
  }
  else
  {
    if(fb->fat12.reserved != 0)
    {
      INFO_PRINT(("Warning -reserved byte #1 not zero!\n"));
      INFO_PRINT(("\t => 0x%02x\n",fb->fat12.reserved));
    }

    if(BOOT_HAVE_BOOTSIG(fb->fat12))
      print_boot_signiature(fb->fat12.bootsig);
    else
      INFO_PRINT(("No boot signiature.\n"));
  }

  return(0);
}

