#ifndef __FAT_INTERNAL_H__
#define __FAT_INTERNAL_H__

#include "fat.h"
#include <stdint.h>
#include "endian.h"

#define BOOT_RESERVED_SECTORS(fb) SDL_SwapLE16((fb)->reserved_sectors)
#define BOOT_BYTES_PER_SECTOR(fb) SDL_SwapLE16((fb)->bytes_per_sector)
#define BOOT_ROOT_ENTRIES(fb)     SDL_SwapLE16((fb)->root_entries)
#define BOOT_SECTORS(fb)          SDL_SwapLE16((fb)->total_sectors)
#define BOOT_FAT_SECTORS(fb)      SDL_SwapLE16((fb)->fat_sectors)
#define BOOT_HIDDEN_SECTORS(fb)   SDL_SwapLE32((fb)->hidden_sectors)
#define BOOT_SECTORS32(fb)        SDL_SwapLE32((fb)->total_sectors_32)

/* Expects fb.fat12, fb.fat16, or fb.fat32 */
#define BOOT_HAVE_BOOTSIG(fat)    ((fat).bootsig_present==0x29)

#define BOOT32_FAT_SECTORS(fb)    SDL_SwapLE32((fb)->fat32.fat_sectors)
#define BOOT32_EXTFLAGS(fb)       SDL_SwapLE16((fb)->fat32.flags)
#define BOOT32_VERSION(fb)        SDL_SwapLE16((fb)->fat32.version)
#define BOOT32_ROOT_CLUSTER(fb)   SDL_SwapLE32((fb)->fat32.root_cluster)
#define BOOT32_FSINFO_SECTOR(fb)  SDL_SwapLE16((fb)->fat32.fsinfo_sector)
#define BOOT32_BOOT_BACKUP(fb)    SDL_SwapLE16((fb)->fat32.bootbackup_sector)

#define BOOT32_ACTIVE_FAT(fb)     (BOOT32_EXTFLAGS(fb) & 0x07)
#define BOOT32_MIRRORED(fb)       (BOOT32_EXTFLAGS(fb) & 0x80)

#define BOOT_CLUSTERSIZE(fb)      (BOOT_BYTES_PER_SECTOR(fb)*fb->sectors_per_cluster)

typedef enum fat_media_e
{
  FAT_MEDIA_FIXED    =0xf8,
  FAT_MEDIA_REMOVABLE=0xf0,
  FAT_UNKNOWN1       =0xf9,
  FAT_UNKNOWN2       =0xfa,
  FAT_UNKNOWN3       =0xfb,
  FAT_UNKNOWN4       =0xfc,
  FAT_UNKNOWN5       =0xfd,
  FAT_UNKNOWN6       =0xfe,
  FAT_UNKNOWN7       =0xff,
} fat_media_e;

#define BOOT_OEMNAME_DEFAULT "MSWIN4.1"

#pragma pack(push,1)

typedef struct FAT_bootsig
{
  uint32_t volumeid;
  char     volumelabel[11];
  char     systype[8];
} FAT_bootsig;

typedef struct FAT_boot
{
  uint8_t  jmpboot[3];          /*3*/
  char     oemname[8];          /*8*/
  uint16_t bytes_per_sector;    /*2*/
  uint8_t  sectors_per_cluster; /*1*/
  uint16_t reserved_sectors;    /*2*/
  uint8_t  num_fats;            /*1*/
  uint16_t root_entries;        /*2*/
  uint16_t total_sectors;       /*2*/
  uint8_t  media_type;          /*1*/
  uint16_t fat_sectors;         /*2*/
  uint16_t sectors_per_track;   /*2*/
  uint16_t num_heads;           /*2*/
  uint32_t hidden_sectors;      /*4*/
  uint32_t total_sectors_32;    /*4*/
  union
  {
    struct {
      uint8_t  int13_drivenum;
      uint8_t  reserved;
      uint8_t  bootsig_present;
      FAT_bootsig bootsig; 
    } fat12,fat16;

    struct {
      uint32_t fat_sectors;
      uint16_t flags;
      uint16_t version;
      uint32_t root_cluster;
      uint16_t fsinfo_sector;
      uint16_t bootbackup_sector;
      uint8_t  reserved[12];
      uint8_t  int13_drivenum;
      uint8_t  reserved1;
      uint8_t  bootsig_present;
      FAT_bootsig bootsig;
    } fat32;
  };

  uint8_t pad[512-92];
  uint8_t hex_55;
  uint8_t hex_aa;

} FAT_boot;

typedef enum cluster_e
{
  CLUSTER_AVAILABLE,
  CLUSTER_RESERVED,
  CLUSTER_USER,
  CLUSTER_BAD,
  CLUSTER_LAST,
  CLUSTER_ERR=-1
} cluster_e;

#define FAT12_AVAILABLE 0x000
#define FAT12_RESERVED  0x001
#define FAT12_BAD       0xff7
#define FAT12_END1      0xff8
#define FAT12_END2      0xfff

#define FAT16_AVAILABLE 0x0000
#define FAT16_RESERVED  0x0001
#define FAT16_BAD       0xfff7
#define FAT16_END1      0xfff8
#define FAT16_END2      0xffff

#define FAT32_AVAILABLE 0x00000000
#define FAT32_RESERVED  0x00000001
#define FAT32_BAD       0x0ffffff7
#define FAT32_END1      0x0ffffff8
#define FAT32_END2      0x0fffffff

typedef union fat_table
{
  uint32_t fat32[0];
  uint16_t fat16[0];
  struct
  {
    uint32_t fat12_0:12;
    uint32_t fat12_1:12;
    uint32_t fat12_2:12;
    uint32_t fat12_3:12;
    uint32_t fat12_4:12;
    uint32_t fat12_5:12;
    uint32_t fat12_6:12;
    uint32_t fat12_7:12;
  } fat12[0];
} fat_table;

typedef enum fat_de_e
{
  FAT_DE_UNUSED = 0x00,
  FAT_DE_E5H    = 0x05,
  FAT_DE_DOT    = 0x2e,
  FAT_DE_DELETED= 0xe5
} fat_de_e;

typedef enum fat_attr_t
{
  FAT_ATTR_LFN       = 0x0f,
  FAT_ATTR_READONLY  = 0x01,
  FAT_ATTR_SYSTEM    = 0x02,
  FAT_ATTR_HIDDEN    = 0x04,
  FAT_ATTR_VOLUME    = 0x08,
  FAT_ATTR_DIRECTORY = 0x10,
  FAT_ATTR_RESERVED1 = 0x20,
  FAT_ATTR_RESERVED2 = 0x40,
  FAT_ATTR_RESERVED3 = 0x80
} fat_attr_t;

typedef struct FAT_direntry
{
  union
  {
    uint8_t info;
    char    name[8];
  };

  char     ext[3];
  uint8_t  attribute;
  uint8_t  reserved1;
  uint8_t  vfat_ms;
  uint16_t vfat_time;
  uint16_t vfat_date;
  uint16_t vfat_adate;
  union
  {
    uint16_t fat32_cluster_hiword;
    uint16_t reserved2;
  };
  uint16_t utime;
  uint16_t udate;
  uint16_t cluster;
  uint32_t size;
} FAT_direntry;

#pragma pack(pop)

typedef uint32_t cluster_t;
typedef uint32_t sector_t;
typedef uint64_t pos_t;

struct FAT
{
  fat_e      type;
  int        fd;
  int        pagesize;
  FAT_boot  *boot;
  fat_table *table;
  uint8_t    sectorbits;
  uint8_t    clusterbits;
  uint32_t   tablesize;
  sector_t   datastart;
  void      *clusterbuf;
//  uint64_t   written;
};

#ifdef __cplusplus
extern "C" {
#endif

static fat_e FAT_fs_type(const FAT_boot *fb);
static int FAT_boot_sector_valid(const FAT_boot *fb);

#ifdef __cplusplus
}
#endif

#endif/*__FAT_INTERNAL_H__*/
